@@ -22,15 +22,19 @@
 #include "memory-pool.h"                // Pool_desc
 #include "utilities.h"                  // locale_t
 
-#define EMPTY_CONNECTOR "ZZZ"
+
+// Dict may have `#define empty-connector ZZZ` in it.
+#define EMPTY_CONNECTOR "empty-connector"
 #define UNLIMITED_CONNECTORS_WORD ("UNLIMITED-CONNECTORS")
 #define LIMITED_CONNECTORS_WORD ("LENGTH-LIMIT-")
 #define IS_GENERATION(dict) (dict->category != NULL)
 
-/* If the maximum disjunct cost is yet uninitialized, the value defined in the
- * dictionary (or if not defined then DEFAULT_MAX_DISJUNCT_COST) is used. */
+/* If the maximum disjunct cost is not initialized, the value defined
+ * in the dictionary is used. If not defined, then DEFAULT_MAX_DISJUNCT_COST
+ * is used. */
 static const float UNINITIALIZED_MAX_DISJUNCT_COST = -10000.0f;
 static const float DEFAULT_MAX_DISJUNCT_COST = 2.7f;
+
 /* We need some of these as literal strings. */
 #define LG_DISJUNCT_COST                        "max-disjunct-cost"
 #define LG_DICTIONARY_VERSION_NUMBER            "dictionary-version-number"
@@ -40,15 +44,6 @@ static const float DEFAULT_MAX_DISJUNCT_COST = 2.7f;
 typedef struct Afdict_class_struct Afdict_class;
 typedef struct Regex_node_s Regex_node;
 
-typedef struct X_node_struct X_node;
-struct X_node_struct
-{
-	const char * string;       /* the word itself */
-	Exp * exp;
-	X_node *next;
-	const Gword *word;         /* originating Wordgraph word */
-};
-
 /* The regexes are stored as a linked list of the following nodes. */
 struct Regex_node_s
 {
@@ -62,7 +57,6 @@ struct Regex_node_s
 	Regex_node *next;
 	bool neg;        /* Negate the match */
 	int capture_group;  /* Capture group number (-1 if none) for ovector */
-	int ovector[2];  /* 0: start offset; 1: end offset */
 };
 
 static inline Regex_node *regex_new(const char *name, const char *pattern)
@@ -129,10 +123,13 @@ struct Dictionary_s
 	const char * version;
 	const char * locale;    /* Locale name */
 	locale_t     lctype;    /* Locale argument for the *_l() functions */
+
 	int          num_entries;
 	float default_max_disjunct_cost;
 	dfine_s      dfine;    /* Name-value definitions */
 
+	const char * zzz_connector;
+
 	bool         use_unknown_word;
 	bool         unknown_word_defined;
 	bool         left_wall_defined;
@@ -173,10 +170,14 @@ struct Dictionary_s
 #endif
 
 	void (*insert_entry)(Dictionary, Dict_node *, int);
+
+	void (*start_lookup)(Dictionary, Sentence);
+	void (*end_lookup)(Dictionary, Sentence);
 	Dict_node* (*lookup_list)(Dictionary, const char*);
 	Dict_node* (*lookup_wild)(Dictionary, const char*);
 	void (*free_lookup)(Dictionary, Dict_node*);
 	bool (*exists_lookup)(Dictionary, const char*);
+
 	void (*clear_cache)(Dictionary);
 	void (*close)(Dictionary);
 
@@ -194,8 +195,10 @@ struct Dictionary_s
 	Category * category;      /* Word lists - indexed by category number */
 	bool generate_walls;      /* Generate walls too for wildcard words */
 
-	/* Private data elements that come in play only while the
-	 * dictionary is being read, and are not otherwise used.
+	/* Private data elements that come in play only while file-backed
+	 * dictionaries are being read, and are not otherwise used.
+	 * TODO: These should be moved to a private structure, accessible
+	 * only to the file backend.
 	 */
 	const char    * input;
 	const char    * pin;
@@ -217,7 +220,6 @@ bool dictionary_generation_request(const Dictionary);
  * and pretty much no one else. If you are not the tokenizer, you
  * probably don't need these. */
 bool dict_has_word(const Dictionary dict, const char *);
-void add_empty_word(Sentence, X_node *);
 
 static inline const char *subscript_mark_str(void)
 {
