@@ -4,7 +4,7 @@
 /* All rights reserved                                                   */
 /*                                                                       */
 /* Use of the link grammar parsing system is subject to the terms of the */
-/* license set forth in the LICENSE file included with this softwares.   */
+/* license set forth in the LICENSE file included with this software.   */
 /* This license allows free redistribution and use in source and binary  */
 /* forms, with or without modification, subject to certain conditions.   */
 /*                                                                       */
@@ -18,7 +18,7 @@
 
 typedef struct extractor_s extractor_t;
 
-extractor_t* extractor_new(int nwords, unsigned int rand_state);
+extractor_t* extractor_new(Sentence);
 void free_extractor(extractor_t*);
 
 bool build_parse_set(extractor_t*, Sentence,
@@ -27,4 +27,6 @@ bool build_parse_set(extractor_t*, Sentence,
 
 void extract_links(extractor_t*, Linkage);
 
+void mark_used_disjuncts(extractor_t *, bool *);
+
 #endif /* _EXTRACT_LINKS_H */
